import warnings

import numpy as np
from scipy.special import comb

from .warnings import SpectralUnitsWarning

__all__ = ["_warn_no_units", "_test_valid_x_range", "_smoothstep"]


def _warn_no_units():
    warnings.warn(
        "x has no units, assuming x units are inverse microns", SpectralUnitsWarning
    )


def _test_valid_x_range(x, x_range, outname):
    """
    Test if any of the x values are outside of the valid range

    Parameters
    ----------
    x : float array
       wavenumbers in inverse microns

    x_range: 2 floats
       allowed min/max of x

    outname: str
       name of curve for error message
    """
    deltacheck = 1e-6  # delta to allow for small numerical issues
    if np.logical_or(
        np.any(x <= (x_range[0] - deltacheck)), np.any(x >= (x_range[1] + deltacheck))
    ):
        raise ValueError(
            "Input x outside of range defined for "
            + outname
            + " ["
            + str(x_range[0])
            + " <= x <= "
            + str(x_range[1])
            + ", x has units 1/micron]"
        )


def _smoothstep(x, x_min=0, x_max=1, N=1):
    x = np.clip((x - x_min) / (x_max - x_min), 0, 1)

    result = 0
    for n in range(0, N + 1):
        result += comb(N + n, n) * comb(2 * N + 1, N - n) * (-x) ** n

    result *= x ** (N + 1)

    return result



import numpy as np
import astropy.units as u

from dust_extinction.tests.helpers import ave_models, param_ave_models, grain_models


print("V band normalization wavelength = A(l)/A(V) = 1")
print("B band normalization wavelength = A(l)/A(V) = 1 + 1/R(V)")
optwaves = np.arange(0.7, 0.4, -0.01) * u.micron
twave = 1 / 0.55
for cmodel in ave_models + param_ave_models + grain_models:
    cmod = cmodel()
    if (twave > cmod.x_range[0]) & (twave < cmod.x_range[1]):
        mvals = cmod(optwaves)
        nwave = np.interp([1.0], mvals, optwaves)
        nwave2 = np.interp([1.0 + 1 / cmod.Rv], mvals, optwaves)
        Rv = cmod.Rv
        if not isinstance(Rv, float):
            Rv = Rv.value
        print(
            f"| {cmod.__class__.__name__:12.12s} | {nwave.squeeze().value:18.4f} | {nwave2.squeeze().value:27.4f} | {Rv:.2f} |"
        )
        print(
            "+--------------+--------------------+-----------------------------+------+"
        )
    else:
        print(cmod.__class__.__name__, " wave coverage doesn't include V band")

print("")
print("R(V) relations where A(l)/A(V) is the same for Rv = 2.5 & 5.0")

for cmodel in param_ave_models[0:-1]:

    cmod1 = cmodel(Rv=2.5)
    cmod2 = cmodel(Rv=5.0)

    if (twave > cmod1.x_range[0]) & (twave < cmod1.x_range[1]):

        mvals1 = cmod1(optwaves)
        mvals2 = cmod2(optwaves)

        # find the wavelength where the difference in the curves is zero
        diff = mvals1 - mvals2
        nwave = np.interp([0.0], diff, optwaves)
        print(cmod1.__class__.__name__, nwave, cmod1(nwave))


# test the G23 original
cmod1 = cmodel(Rv=2.5, renorm=False)
cmod2 = cmodel(Rv=5.0, renorm=False)

mvals1 = cmod1(optwaves)
mvals2 = cmod2(optwaves)

# find the wavelength where the difference in the curves is zero
diff = mvals1 - mvals2
nwave = np.interp([0.0], diff, optwaves)
print(cmod1.__class__.__name__, "orig", nwave, cmod1(nwave))



import numpy as np
import pytest

import astropy.units as u
from astropy.modeling.fitting import LevMarLSQFitter

from ..averages import G03_LMCAvg
from ..shapes import FM90


def get_elvebv_cor_vals():
    # testing wavenumbers
    x = np.array([3.5, 4.0, 4.5, 5.0, 6.0, 7.0, 9.0, 10.0])

    # add units
    x = x / u.micron

    # correct values
    # fmt: off
    cor_vals = np.array([2.995507, 4.187955, 6.70251, 5.723752, 4.757428,
                         5.490276, 9.285265, 12.462183])
    # fmt: on

    return (x, cor_vals)


def test_extinction_FM90_values():
    # get the correct values
    x, cor_vals = get_elvebv_cor_vals()

    # initialize extinction model
    tmodel = FM90()

    # test
    np.testing.assert_allclose(tmodel(x), cor_vals)


x_vals, axav_vals = get_elvebv_cor_vals()
test_vals = zip(x_vals, axav_vals)


@pytest.mark.parametrize("xtest_vals", test_vals)
def test_extinction_FM90_single_values(xtest_vals):
    x, cor_val = xtest_vals

    # initialize extinction model
    tmodel = FM90()

    # test
    np.testing.assert_allclose(tmodel(x), cor_val)


def test_FM90_fitting():

    # get an observed extinction curve to fit
    g03_model = G03_LMCAvg()

    x = g03_model.obsdata_x
    # convert to E(x-V)/E(B0V)
    y = (g03_model.obsdata_axav - 1.0) * g03_model.Rv
    # only fit the UV portion (FM90 only valid in UV)
    (gindxs,) = np.where(x > 3.125)

    fm90_init = FM90()
    fit = LevMarLSQFitter()
    g03_fit = fit(fm90_init, x[gindxs], y[gindxs])
    fit_vals = [
        g03_fit.C1.value,
        g03_fit.C2.value,
        g03_fit.C3.value,
        g03_fit.C4.value,
        g03_fit.xo.value,
        g03_fit.gamma.value,
    ]

    # fmt: off
    good_vals = np.array([-0.941674, 1.013711, 2.725373, 0.301217,
                          4.589078, 0.948576])
    # fmt: on

    np.testing.assert_allclose(good_vals, fit_vals, rtol=1e-5)



import numpy as np
import pytest

import astropy.units as u
from astropy.modeling.fitting import LevMarLSQFitter

from dust_extinction.shapes import P92


def get_axav_cor_vals():

    # Milky Way observed extinction as tabulated by Pei (1992)
    # fmt: off
    MW_x = [0.21, 0.29, 0.45, 0.61, 0.80, 1.11, 1.43, 1.82,
            2.27, 2.50, 2.91, 3.65, 4.00, 4.17, 4.35, 4.57, 4.76,
            5.00, 5.26, 5.56, 5.88, 6.25, 6.71, 7.18, 7.60,
            8.00, 8.50, 9.00, 9.50, 10.00]
    MW_x = np.array(MW_x)
    MW_exvebv = [-3.02, -2.91, -2.76, -2.58, -2.23, -1.60, -0.78, 0.00,
                 1.00, 1.30, 1.80, 3.10, 4.19, 4.90, 5.77, 6.57, 6.23,
                 5.52, 4.90, 4.65, 4.60, 4.73, 4.99, 5.36, 5.91,
                 6.55, 7.45, 8.45, 9.80, 11.30]
    # fmt: on

    MW_exvebv = np.array(MW_exvebv)
    Rv = 3.08
    MW_axav = MW_exvebv / Rv + 1.0

    # add units
    x = MW_x / u.micron

    # correct values
    cor_vals = MW_axav

    return (x, cor_vals)


def test_extinction_P92_values():
    # get the correct values
    x, cor_vals = get_axav_cor_vals()

    # initialize extinction model
    tmodel = P92()

    # test
    np.testing.assert_allclose(tmodel(x), cor_vals, rtol=0.25, atol=0.01)


x_vals, axav_vals = get_axav_cor_vals()
test_vals = zip(x_vals, axav_vals)


@pytest.mark.parametrize("xtest_vals", test_vals)
def test_extinction_P92_single_values(xtest_vals):
    x, cor_val = xtest_vals

    # initialize extinction model
    tmodel = P92()

    # test
    np.testing.assert_allclose(tmodel(x), cor_val, rtol=0.25, atol=0.01)


def test_P92_fitting():

    # get an observed extinction curve to fit
    x_quant, y = get_axav_cor_vals()

    x = x_quant.value

    # change from defaults to make the best fit harder to find
    p92_init = P92()

    fit = LevMarLSQFitter()
    # accuracy set to avoid warning the fit may have failed
    p92_fit = fit(p92_init, x, y, acc=1e-2)

    fit_vals = p92_fit._parameters

    # fmt: off
    good_vals = [220.43887482191883, 0.04772654363744676, 89.13955305851606, 2.0,
                 17.53404746789474, 0.072502973523927, 1.6026274827983342,
                 5.454549320255999, 0.05379254892041606, 0.2187391000979195,
                 -1.9501411144824525, 2.0, 0.0, 13.0, 504.5387984216088, 2.0,
                 0.0, 15.0, -5192.423852001501, 2.0, 0.0, 30.0,
                 -10264.367792916015, 2.0]
    # fmt: on

    np.testing.assert_allclose(good_vals, fit_vals)



import argparse
import numpy as np
import matplotlib.pyplot as plt

import astropy.units as u
from dust_extinction.parameter_averages import G23

def rect(x,y,w,h,c):
    ax = plt.gca()
    polygon = plt.Rectangle((x,y),w,h,color=c)
    ax.add_patch(polygon)

def rainbow_fill(X,Y, cmap=plt.get_cmap("jet")):
    plt.plot(X,Y,lw=0)  # Plot so the axes scale correctly

    dx = X[1]-X[0]
    N  = float(X.size)

    for n, (x,y) in enumerate(zip(X,Y)):
        color = cmap(n/N)
        rect(x,0,dx,y,color)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--png", help="save figure as a pdf file", action="store_true")
    args = parser.parse_args()

    # plotting setup for easier to read plots
    fontsize = 18
    font = {"size": fontsize}
    plt.rc("font", **font)
    plt.rc("lines", linewidth=1)
    plt.rc("axes", linewidth=2)
    plt.rc("xtick.major", width=2)
    plt.rc("xtick.minor", width=2)
    plt.rc("ytick.major", width=2)
    plt.rc("ytick.minor", width=2)

    # create the plot
    fig, ax = plt.subplots(figsize=(8, 8))

    ext = G23()

    lam = np.logspace(np.log10(0.0912), np.log10(30.0), num=1000) * u.micron

    ext1 = ext(lam)
    ext2 = np.log10(ext1)

    weight = np.log10(lam.value) / (np.log10(30.0) - np.log10(0.0912))

    #ax.plot(lam, ext1, "k-", lw=2)
    #ax.plot(lam, ext2 + 1.5, "b-", lw=2)
    mergeext = ext1 + (weight * (ext2 + 0.5)) + 1.0
    # im = ax.plot(lam, mergeext, "g-", lw=2)

    #patch = patches.Circle((1.0, 3.0), radius=3, transform=ax.transData)
    #im[0].set_clip_path(patch)

    rainbow_fill(lam.value, mergeext)

    ax.axis('off')

    ax.set_xscale("log")

    # show the figure or save it to a pdf file
    if args.png:
        fig.savefig("dust_extinction_log.png")
        plt.close()
    else:
        plt.show()


import numpy as np
import astropy.units as u
from astropy.table import QTable
from dust_extinction.parameter_averages import G23
import matplotlib.pyplot as plt

if __name__ == '__main__':

    Rv_range = [2.0, 6.51]
    n_Rvs = int(np.round((Rv_range[1] - Rv_range[0]) / 0.1) + 1.0)
    print("# Rvs", n_Rvs)
    Rvs = np.array(range(n_Rvs)) * 0.1 + Rv_range[0]
    print(Rvs)
    # Rvs = [2.3, 2.7, 3.1, 3.5, 4.0, 4.5, 5.0, 5.6]
    waves = np.logspace(np.log10(0.0912), np.log10(32.), num=1000) * u.micron

    otab = QTable()
    otab.meta["keywords"] = {}
    otab.meta["keywords"]["origin"] = {"value": "dust_extinction python package"}
    otab.meta["keywords"]["model"] = {"value": "G23"}
    otab.meta["keywords"]["author"] = {"value": "Karl D. Gordon"}
    otab["wavelength"] = waves
    for cRv in Rvs:
        cmod = G23()
        # override Rv_range to allow extrapolation - here be dragons
        #  Above Rv=6.3, FUV extinction becomes lower at shorter wavelengths
        #  Below Rv=2.2, MIR before 10um silicates goes unrealistically small
        cmod.Rv_range = Rv_range
        cmod.Rv = cRv
        otab.meta["keywords"]["Rv"] = {"value": cRv}
        otab["AlambdaAv"] = cmod(waves)
        otab.write(f"dust_extinction_G23_Rv_{cRv:.1f}.ipac", format="ascii.ipac", overwrite=True)

        plt.plot(otab["wavelength"], otab["AlambdaAv"])

    plt.xscale("log")
    plt.yscale("log")
    plt.show()
